#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi
import troposphere as t
import yaml
from troposphere import codebuild, events, iam, s3

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class PrepareC7NHubAccountTask(tasks.TaskWithReferenceAndCommonParameters):
    custodian_region = luigi.Parameter()
    c7n_version = luigi.Parameter()
    c7n_org_version = luigi.Parameter()
    organization = luigi.Parameter()
    role_name = luigi.Parameter()
    role_path = luigi.Parameter()
    schedule_expression = luigi.Parameter()

    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
        }

    def run(self):
        # TODO move to troposphere
        template = f"""
  eventbuspolicy:
    Type: AWS::Events::EventBusPolicy
    Properties:
      Condition:
        Key: aws:PrincipalOrgID
        Type: StringEquals
        Value: {self.organization}
      Action: events:PutEvents
      Principal: '*'
      StatementId: OrganizationAccounts
    """
        tpl = t.Template()
        tpl.description = "event bus template for c7n created by service catalog puppet"

        tpl.add_resource(
            iam.Role(
                "C7NRunRole",
                RoleName="C7NRunRole",
                AssumeRolePolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {"Service": ["codebuild.amazonaws.com"]},
                        }
                    ],
                },
                Policies=[
                    iam.Policy(
                        PolicyName="C7NRunRoleActions",
                        PolicyDocument={
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": [
                                        "sts:AssumeRole",
                                    ],
                                    "Resource": t.Sub(
                                        "arn:${AWS::Partition}:iam::*:role"
                                        + self.role_path
                                        + self.role_name
                                    ),
                                    "Effect": "Allow",
                                },
                                {
                                    "Action": [
                                        "logs:CreateLogStream",
                                        "logs:CreateLogGroup",
                                        "logs:PutLogEvents",
                                    ],
                                    "Resource": t.Sub(
                                        "arn:${AWS::Partition}:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/codebuild/servicecatalog-puppet-deploy-c7n:log-stream:*"
                                    ),
                                    "Effect": "Allow",
                                },
                                {
                                    "Action": [
                                        "logs:GetLogEvents",
                                    ],
                                    "Resource": t.Sub(
                                        "arn:${AWS::Partition}:logs:${AWS::Region}:${AWS::AccountId}:log-group:/aws/codebuild/servicecatalog-puppet-deploy-c7n:log-stream:*"
                                    ),
                                    "Effect": "Allow",
                                },
                                {
                                    "Action": [
                                        "ssm:GetParameters",
                                    ],
                                    "Resource": [
                                        t.Sub(
                                            "arn:${AWS::Partition}:ssm:${AWS::Region}:${AWS::AccountId}:parameter/servicecatalog-puppet/aws-c7n-lambdas/REGIONS"
                                        ),
                                        t.Sub(
                                            "arn:${AWS::Partition}:ssm:${AWS::Region}:${AWS::AccountId}:parameter/servicecatalog-puppet/aws-c7n-lambdas/CUSTODIAN_ROLE_ARN"
                                        ),
                                    ],
                                    "Effect": "Allow",
                                },
                                {
                                    "Action": [
                                        "s3:GetObject",
                                        "s3:GetObjectVersion",
                                    ],
                                    "Resource": [
                                        t.Sub(
                                            "arn:aws:s3:::sc-puppet-c7n-artifacts-${AWS::AccountId}-${AWS::Region}/latest"
                                        ),
                                    ],
                                    "Effect": "Allow",
                                },
                            ],
                        },
                    )
                ],
                Path=config.get_puppet_role_path(),
            )
        )

        codebuild_target = t.Sub(
            "arn:${AWS::Partition}:codebuild:${AWS::Region}:${AWS::AccountId}:project/servicecatalog-puppet-deploy-c7n"
        )

        tpl.add_resource(
            iam.Role(
                "C7NEventRuleRunRole",
                RoleName="C7NEventRuleRunRole",
                AssumeRolePolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {"Service": ["events.amazonaws.com"]},
                        }
                    ],
                },
                Policies=[
                    iam.Policy(
                        PolicyName="C7NEventRuleRunRoleActions",
                        PolicyDocument={
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Action": [
                                        "codebuild:StartBuild",
                                    ],
                                    "Resource": [
                                        codebuild_target,
                                    ],
                                    "Effect": "Allow",
                                },
                            ],
                        },
                    )
                ],
                Path=config.get_puppet_role_path(),
            )
        )

        tpl.add_resource(
            s3.Bucket(
                "c7nPoliciesBucket",
                BucketName=t.Sub(
                    "sc-puppet-c7n-artifacts-${AWS::AccountId}-" + self.custodian_region
                ),
                VersioningConfiguration=s3.VersioningConfiguration(Status="Enabled"),
                BucketEncryption=s3.BucketEncryption(
                    ServerSideEncryptionConfiguration=[
                        s3.ServerSideEncryptionRule(
                            ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                                SSEAlgorithm="AES256"
                            )
                        )
                    ]
                ),
                PublicAccessBlockConfiguration=s3.PublicAccessBlockConfiguration(
                    BlockPublicAcls=True,
                    BlockPublicPolicy=True,
                    IgnorePublicAcls=True,
                    RestrictPublicBuckets=True,
                ),
                Tags=t.Tags({"ServiceCatalogPuppet:Actor": "Framework"}),
            )
        )

        if self.c7n_org_version == "":
            build_spec = dict(
                version=0.2,
                phases=dict(
                    install={"commands": ["pip install c7n==${C7N_VERSION}"]},
                    build={
                        "commands": [
                            "aws s3 cp s3://sc-puppet-c7n-artifacts-${ACCOUNT_ID}-${REGION}/latest latest.zip",
                            "unzip latest.zip",
                            "for REGION in ${REGIONS}; do custodian run -s output/logs -r ${REGION} --assume ${CUSTODIAN_ROLE_ARN} policies.yaml; done",
                        ]
                    },
                ),
            )
        else:
            build_spec = dict(
                version=0.2,
                phases=dict(
                    install={"commands": ["pip install c7n-org==${C7N_ORG_VERSION}"]},
                    build={
                        "commands": [
                            "aws s3 cp s3://sc-puppet-c7n-artifacts-${ACCOUNT_ID}-${REGION}/latest latest.zip",
                            "unzip latest.zip",
                            "c7n-org run -c accounts.yaml -s output -u policies.yaml",
                        ]
                    },
                ),
            )

        tpl.add_resource(
            codebuild.Project(
                "DeployC7N",
                Name="servicecatalog-puppet-deploy-c7n",
                ServiceRole=t.GetAtt("C7NRunRole", "Arn"),
                Tags=t.Tags.from_dict(**{"ServiceCatalogPuppet:Actor": "Framework"}),
                Artifacts=codebuild.Artifacts(Type="NO_ARTIFACTS"),
                TimeoutInMinutes=60 * 8,
                Environment=codebuild.Environment(
                    ComputeType="BUILD_GENERAL1_SMALL",
                    Image=constants.CODEBUILD_DEFAULT_IMAGE,
                    Type="LINUX_CONTAINER",
                    EnvironmentVariables=[
                        {
                            "Type": "PLAINTEXT",
                            "Name": "C7N_VERSION",
                            "Value": self.c7n_version,
                        },
                        {
                            "Type": "PLAINTEXT",
                            "Name": "C7N_ORG_VERSION",
                            "Value": self.c7n_org_version,
                        },
                        {
                            "Type": "PLAINTEXT",
                            "Name": "ACCOUNT_ID",
                            "Value": t.Ref("AWS::AccountId"),
                        },
                        {
                            "Type": "PLAINTEXT",
                            "Name": "REGION",
                            "Value": t.Ref("AWS::Region"),
                        },
                        {
                            "Type": "PARAMETER_STORE",
                            "Name": "REGIONS",
                            "Value": "/servicecatalog-puppet/aws-c7n-lambdas/REGIONS",
                        },
                        {
                            "Type": "PARAMETER_STORE",
                            "Name": "CUSTODIAN_ROLE_ARN",
                            "Value": "/servicecatalog-puppet/aws-c7n-lambdas/CUSTODIAN_ROLE_ARN",
                        },
                    ],
                ),
                Source=codebuild.Source(
                    BuildSpec=yaml.safe_dump(build_spec),
                    Type="NO_SOURCE",
                ),
                Description="Run c7n",
            )
        )

        tpl.add_resource(
            events.Rule(
                "Runner",
                Name="servicecatalog-puppet-c7n-scheduled-runner",
                RoleArn=t.GetAtt("C7NEventRuleRunRole", "Arn"),
                ScheduleExpression=self.schedule_expression,
                Targets=[
                    events.Target(
                        Id="RunCodeBuild",
                        Arn=codebuild_target,
                        RoleArn=t.GetAtt("C7NEventRuleRunRole", "Arn"),
                    )
                ],
                State="ENABLED" if self.schedule_expression != "" else "DISABLED",
            )
        )

        template = tpl.to_yaml() + template

        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-c7n-eventbus",
                Capabilities=["CAPABILITY_NAMED_IAM"],
                TemplateBody=template,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
        self.write_output(
            dict(
                c7n_account_id=self.account_id,
            )
        )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import aws, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class TerminateSpokeLocalPortfolioTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            if self.puppet_account_id == self.account_id:
                self.write_empty_output()
            else:
                portfolio = aws.find_portfolio(servicecatalog, self.portfolio)
                if portfolio is False:
                    self.write_empty_output()
                else:
                    servicecatalog.delete_portfolio(Id=portfolio.get("Id"))
                    self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi
import troposphere as t
from troposphere import iam

from servicecatalog_puppet import config, constants
from servicecatalog_puppet.workflow.dependencies import tasks


class CreateCustodianRoleTask(tasks.TaskWithReferenceAndCommonParameters):
    c7n_account_id = luigi.Parameter()
    role_name = luigi.Parameter()
    role_path = luigi.Parameter()
    role_managed_policy_arns = luigi.ListParameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
        }

    def run(self):
        c7n_account_id = self.c7n_account_id
        role_name = self.role_name
        role_path = self.role_path
        role_managed_policy_arns = self.role_managed_policy_arns

        tpl = t.Template()
        tpl.description = (
            "custodian role template for c7n created by service catalog puppet"
        )
        tpl.add_resource(
            iam.Role(
                role_name,
                RoleName=role_name,
                ManagedPolicyArns=[t.Sub(arn) for arn in role_managed_policy_arns],
                Path=role_path,
                AssumeRolePolicyDocument={
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {"Service": ["lambda.amazonaws.com"]},
                        },
                        {
                            "Action": ["sts:AssumeRole"],
                            "Effect": "Allow",
                            "Principal": {
                                "AWS": t.Sub(
                                    "arn:${AWS::Partition}:iam::"
                                    + c7n_account_id
                                    + ":root"
                                )
                            },
                        },
                    ],
                },
            )
        )
        template = tpl.to_yaml()
        with self.spoke_regional_client("cloudformation") as cloudformation:
            cloudformation.create_or_update(
                Capabilities=["CAPABILITY_NAMED_IAM"],
                ShouldUseChangeSets=False,
                StackName="servicecatalog-puppet-c7n-custodian",
                TemplateBody=template,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
        self.write_empty_output()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import queue
import time
import traceback

from servicecatalog_puppet import serialisation_utils
from servicecatalog_puppet.serialisation_utils import unwrap
from servicecatalog_puppet.waluigi.constants import COMPLETED, ERRORED
from servicecatalog_puppet.waluigi.dag_utils import logger
from servicecatalog_puppet.waluigi.locks.external import (
    are_resources_are_free_for_task,
    lock_resources_for_task,
    unlock_resources_for_task,
)
from servicecatalog_puppet.workflow.dependencies import task_factory


def worker_task(
    thread_name,
    lock,
    task_queue,
    results_queue,
    task_processing_time_queue,
    task_trace_queue,
    control_queue,
    control_event,
    manifest_files_path,
    manifest_task_reference_file_path,
    puppet_account_id,
    resources_file_path,
    all_tasks,
    resources,
    tasks_to_run,
):
    logger.info(f"starting up")
    should_be_running = True
    if control_event:
        should_be_running = not control_event.is_set()
    while should_be_running:
        if control_event:
            time.sleep(0.1)
        try:
            task_reference = task_queue.get(timeout=5)
        except queue.Empty:
            if control_event:
                should_be_running = not control_event.is_set()
            continue
        else:
            result = False
            while not result:
                # print(
                #     f"{pid} Worker received {task_reference} waiting for lock",
                #     flush=True,
                # )
                task_parameters = all_tasks.get(task_reference)
                # print(
                #     f"{pid} Worker received {task_reference} waiting for lock and the task is {task_parameters}",
                #     flush=True,
                # )

                with lock:
                    # print(f"{pid} Worker {task_reference} got the lock", flush=True)
                    (
                        resources_are_free,
                        resources_in_use,
                    ) = are_resources_are_free_for_task(
                        task_parameters, resources_file_path
                    )
                    # print(
                    #     f"{pid} Worker {task_reference} resources_are_free: {resources_are_free}",
                    #     flush=True,
                    # )
                    if resources_are_free:
                        lock_resources_for_task(
                            task_reference,
                            task_parameters,
                            resources_in_use,
                            resources_file_path,
                        )
                        # print(f"{pid} Worker {task_reference} locked", flush=True)

                if resources_are_free:
                    # print(f"{pid} Worker about to run {task_reference}", flush=True)
                    task = task_factory.create(
                        manifest_files_path=manifest_files_path,
                        manifest_task_reference_file_path=manifest_task_reference_file_path,
                        puppet_account_id=puppet_account_id,
                        parameters_to_use=task_parameters,
                    )
                    logger.info(f"executing task: {task_reference}")
                    task.on_task_start()
                    start = time.time()
                    task_type, task_details = task.get_processing_time_details()
                    task_trace_queue.put(
                        (start, task_type, task_details, True, thread_name),
                    )
                    try:
                        task.execute()
                    except Exception as e:
                        end = time.time()
                        duration = end - start
                        result = ERRORED
                        logger.error(
                            f"executed task [failure]: {task_reference} failures: {e}"
                        )
                        try:
                            logger.error(f"---- START OF ERROR----")
                            logger.error(f"Task {task_type}:")
                            for l in serialisation_utils.dump(
                                unwrap(task_details)
                            ).split("\n"):
                                logger.error(l)
                            for l in traceback.format_exception(
                                e,
                                value=e,
                                tb=e.__traceback__,
                            ):
                                for sl in l.split("\n"):
                                    logger.error(f"{sl}")
                            logger.error(f"---- END OF ERROR ----")
                        except Exception as e2:
                            logger.error(
                                f"Exception raised: {e2}  whilst logging exection: {e}"
                            )
                        task.on_task_failure(e, duration)
                    else:
                        end = time.time()
                        duration = end - start
                        result = COMPLETED
                        task.on_task_success(duration)

                    task_processing_time_queue.put(
                        (
                            duration,
                            task_type,
                            task_details,
                        ),
                    )
                    task_trace_queue.put(
                        (end, task_type, task_details, False, thread_name),
                    )

                    # print(f"{pid} Worker {task_reference} waiting for lock to unlock resources", flush=True)
                    with lock:
                        logger.info(
                            f"executed task [success]: {task_reference} got lock to unlock resources"
                        )
                        unlock_resources_for_task(task_parameters, resources_file_path)
                        results_queue.put((task_reference, result))
                        time.sleep(0.1)
                else:
                    time.sleep(0.01)

        # time.sleep(10)
        if control_event:
            should_be_running = not control_event.is_set()
    logger.info(f"shutting down")



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0
from servicecatalog_puppet import (
    config,
    constants,
    task_reference_constants,
    utils,
)
from servicecatalog_puppet.commands.task_reference_helpers.generators import portfolios
from servicecatalog_puppet.workflow.portfolio.associations import association_utils


def get_spoke_local_portfolio_common_args(
    task_to_add, all_tasks_task_reference, extra_dependencies_by_reference=[]
):
    return dict(
        status=task_to_add.get("status"),
        account_id=task_to_add.get("account_id"),
        region=task_to_add.get("region"),
        portfolio=task_to_add.get("portfolio"),
        execution=task_to_add.get("execution"),
        dependencies_by_reference=[all_tasks_task_reference]
        + extra_dependencies_by_reference,
        portfolio_task_reference=all_tasks_task_reference,
    )


def handle_spoke_local_portfolios(
    all_tasks,
    all_tasks_task_reference,
    item_name,
    puppet_account_id,
    section_name,
    task_reference,
    task_to_add,
):
    is_sharing_with_puppet_account = task_to_add.get("account_id") == puppet_account_id

    if task_to_add.get("status") == constants.TERMINATED:
        deps = list()
        # DELETE THE ASSOCIATION IF IT EXISTS
        if task_to_add.get("associations"):
            shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            ref = f"portfolio_associations-{shared_ref}"
            deps.append(ref)
            if not all_tasks.get(ref):
                all_tasks[ref] = {
                    "status": task_to_add.get("status"),
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "dependencies_by_reference": [constants.CREATE_POLICIES],
                    "task_reference": ref,
                    "spoke_local_portfolio_name": item_name,
                    "section_name": constants.PORTFOLIO_ASSOCIATIONS,
                    "associations": task_to_add.get("associations"),
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
                task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )
        # DELETE THE LAUNCH CONSTRAINTS IF IT EXISTS
        if task_to_add.get("launch_constraints"):
            shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            ref = f"launch_constraints-{shared_ref}"
            deps.append(ref)
            if not all_tasks.get(ref):
                all_tasks[ref] = {
                    "status": task_to_add.get("status"),
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "dependencies_by_reference": [constants.CREATE_POLICIES],
                    "task_reference": ref,
                    "section_name": constants.PORTFOLIO_CONSTRAINTS_LAUNCH,
                    "spoke_local_portfolio_name": item_name,
                    "launch_constraints": task_to_add["launch_constraints"],
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
                task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )
        # DELETE THE RESOURCE UPDATE CONSTRAINTS IF IT EXISTS
        if task_to_add.get("resource_update_constraints"):
            shared_ref = f"{section_name}-{item_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            ref = f"resource_update_constraints-{shared_ref}"
            deps.append(ref)
            if not all_tasks.get(ref):
                all_tasks[ref] = {
                    "status": task_to_add.get("status"),
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "dependencies_by_reference": [constants.CREATE_POLICIES],
                    "task_reference": ref,
                    "section_name": constants.PORTFOLIO_CONSTRAINTS_RESOURCE_UPDATE,
                    "spoke_local_portfolio_name": item_name,
                    "resource_update_constraints": task_to_add[
                        "resource_update_constraints"
                    ],
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES].update(
                task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS].update(
                task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )
        # GET THE SPOKE LOCAL PORTFOLIO SO WE CAN DELETE THE ASSOCIATIONS
        spoke_portfolio_ref = f"{constants.PORTFOLIO_LOCAL}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        if not all_tasks.get(spoke_portfolio_ref):
            all_tasks[spoke_portfolio_ref] = {
                "puppet_account_id": puppet_account_id,
                "task_reference": spoke_portfolio_ref,
                "dependencies_by_reference": [constants.CREATE_POLICIES],
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "status": task_to_add.get("status"),
                "execution": task_to_add.get("execution"),
                "section_name": constants.PORTFOLIO_LOCAL,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        all_tasks[spoke_portfolio_ref][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

        if not is_sharing_with_puppet_account:
            # GET THE SPOKE PRODUCTS AND VERSIONS SO WE CAN DISASSOCIATE THEM
            spoke_portfolio_all_products_and_versions_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_all_products_and_versions_ref):
                all_tasks[spoke_portfolio_all_products_and_versions_ref] = {
                    "execution": task_to_add.get("execution"),
                    "puppet_account_id": puppet_account_id,
                    "task_reference": spoke_portfolio_all_products_and_versions_ref,
                    "dependencies_by_reference": [spoke_portfolio_ref],
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "portfolio_task_reference": spoke_portfolio_ref,
                    "section_name": constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            disassociate_portfolio_all_products_and_versions_ref = f"{constants.PORTFOLIO_DISASSOCIATE_ALL_PRODUCTS_AND_THEIR_VERSIONS}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(disassociate_portfolio_all_products_and_versions_ref):
                all_tasks[disassociate_portfolio_all_products_and_versions_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": disassociate_portfolio_all_products_and_versions_ref,
                    "dependencies_by_reference": [
                        spoke_portfolio_all_products_and_versions_ref,
                        spoke_portfolio_ref,
                        constants.CREATE_POLICIES,
                    ],
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "portfolio_task_reference": spoke_portfolio_ref,
                    "section_name": constants.PORTFOLIO_DISASSOCIATE_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[disassociate_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[disassociate_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[disassociate_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            deps.append(disassociate_portfolio_all_products_and_versions_ref)

        # DELETE THE SPOKE LOCAL PORTFOLIO ASSOCIATION
        spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
        if not all_tasks.get(spoke_portfolio_puppet_association_ref):
            all_tasks[spoke_portfolio_puppet_association_ref] = {
                "status": task_to_add.get("status"),
                "puppet_account_id": puppet_account_id,
                "task_reference": spoke_portfolio_puppet_association_ref,
                "portfolio_task_reference": spoke_portfolio_ref,
                "dependencies_by_reference": [
                    spoke_portfolio_ref,
                    constants.CREATE_POLICIES,
                ],
                "account_id": task_to_add.get("account_id"),
                "region": task_to_add.get("region"),
                "portfolio": task_to_add.get("portfolio"),
                "execution": task_to_add.get("execution"),
                "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
                task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
            }
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_SECTION_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_ITEM_NAMES
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
        all_tasks[spoke_portfolio_puppet_association_ref][
            task_reference_constants.MANIFEST_ACCOUNT_IDS
        ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
        deps.append(spoke_portfolio_puppet_association_ref)

        task_to_add["dependencies_by_reference"].extend(deps)

    else:
        dependencies_for_constraints = list()
        if is_sharing_with_puppet_account:
            # CREATE SPOKE ASSOCIATIONS
            spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_puppet_association_ref):
                all_tasks[spoke_portfolio_puppet_association_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": spoke_portfolio_puppet_association_ref,
                    "portfolio_task_reference": all_tasks_task_reference,
                    "dependencies_by_reference": [
                        all_tasks_task_reference,
                        constants.CREATE_POLICIES,
                    ],
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # GET THE SPOKE PRODUCTS AND VERSIONS - USE THE AFTER ONE AS NO PRODUCTS WILL BE COPIED WHEN SHARING WITH PUPPET ACCOUNT
            spoke_portfolio_all_products_and_versions_after_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-after-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_all_products_and_versions_after_ref):
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        [spoke_portfolio_puppet_association_ref],
                    ),
                    task_reference=spoke_portfolio_all_products_and_versions_after_ref,
                    section_name=constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                )
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ] = dict()
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
            dependencies_for_constraints.append(
                spoke_portfolio_all_products_and_versions_after_ref
            )

        else:
            # GET THE HUB PORTFOLIO
            hub_portfolio_ref = f"{constants.PORTFOLIO_LOCAL}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(hub_portfolio_ref):
                all_tasks[hub_portfolio_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": hub_portfolio_ref,
                    "dependencies_by_reference": [],
                    "account_id": puppet_account_id,
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "status": task_to_add.get("status"),
                    "execution": task_to_add.get("execution"),
                    "section_name": constants.PORTFOLIO_LOCAL,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[hub_portfolio_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[hub_portfolio_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[hub_portfolio_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
            all_tasks[all_tasks_task_reference][
                "portfolio_task_reference"
            ] = hub_portfolio_ref
            all_tasks[all_tasks_task_reference]["dependencies_by_reference"].append(
                hub_portfolio_ref
            )

            # CREATE THE HUB ASSOCIATIONS
            hub_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(hub_portfolio_puppet_association_ref):
                all_tasks[hub_portfolio_puppet_association_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": hub_portfolio_puppet_association_ref,
                    "portfolio_task_reference": hub_portfolio_ref,
                    "dependencies_by_reference": [
                        hub_portfolio_ref,
                        constants.CREATE_POLICIES,
                    ],
                    "account_id": puppet_account_id,
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,  # TODO test in with a new spoke local
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[hub_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[hub_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[hub_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # SHARE THE PORTFOLIO
            sharing_mode = task_to_add.get(
                "sharing_mode", config.get_global_sharing_mode_default()
            )
            if sharing_mode == constants.SHARING_MODE_ACCOUNT:
                share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
                sharing_type = "ACCOUNT"
            elif sharing_mode == constants.SHARING_MODE_AWS_ORGANIZATIONS:
                share_and_accept_ref = f"portfolio_share_and_accept-{task_to_add.get('ou')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
                sharing_type = (
                    "ORGANIZATIONAL_UNIT"
                    if task_to_add.get("ou")[0:3] == "ou-"
                    else "ORGANIZATION"
                )
            else:
                raise Exception(f"Unknown sharing mode: {sharing_mode}")

            describe_portfolio_shares_task_ref = (
                portfolios.get_or_create_describe_portfolio_shares_task_ref(
                    all_tasks,
                    puppet_account_id,
                    sharing_type,
                    hub_portfolio_ref,
                    task_to_add,
                )
            )

            if not all_tasks.get(share_and_accept_ref):
                all_tasks[share_and_accept_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "task_reference": share_and_accept_ref,
                    "share_tag_options": task_to_add.get("share_tag_options"),
                    "share_principals": task_to_add.get("share_principals"),
                    "dependencies_by_reference": [
                        hub_portfolio_ref,
                        constants.CREATE_POLICIES,
                        describe_portfolio_shares_task_ref,
                    ],
                    "describe_portfolio_shares_task_ref": describe_portfolio_shares_task_ref,
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "portfolio_task_reference": hub_portfolio_ref,
                    "section_name": f"portfolio-share-and-accept-{sharing_mode.lower()}",
                    "ou_to_share_with": task_to_add.get("ou"),
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[share_and_accept_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[share_and_accept_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[share_and_accept_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # GET THE HUB PRODUCTS AND VERSIONS SO WE KNOW WHAT NEEDS TO BE COPIED OR IMPORTED
            hub_portfolio_all_products_and_versions_before_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-before-{puppet_account_id}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(hub_portfolio_all_products_and_versions_before_ref):
                all_tasks[hub_portfolio_all_products_and_versions_before_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": hub_portfolio_all_products_and_versions_before_ref,
                    "dependencies_by_reference": [
                        hub_portfolio_ref,
                        hub_portfolio_puppet_association_ref,  # TODO reduce this down to one ?
                    ],
                    "portfolio_task_reference": hub_portfolio_ref,
                    "account_id": puppet_account_id,
                    "region": task_to_add.get("region"),
                    "execution": task_to_add.get("execution"),
                    "section_name": constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[hub_portfolio_all_products_and_versions_before_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[hub_portfolio_all_products_and_versions_before_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[hub_portfolio_all_products_and_versions_before_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # CREATE SPOKE ASSOCIATIONS
            spoke_portfolio_puppet_association_ref = f"{constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_puppet_association_ref):
                all_tasks[spoke_portfolio_puppet_association_ref] = {
                    "puppet_account_id": puppet_account_id,
                    "task_reference": spoke_portfolio_puppet_association_ref,
                    "portfolio_task_reference": all_tasks_task_reference,
                    "dependencies_by_reference": [
                        all_tasks_task_reference,
                        constants.CREATE_POLICIES,
                        share_and_accept_ref,
                    ],
                    "account_id": task_to_add.get("account_id"),
                    "region": task_to_add.get("region"),
                    "portfolio": task_to_add.get("portfolio"),
                    "execution": task_to_add.get("execution"),
                    "section_name": constants.PORTFOLIO_PUPPET_ROLE_ASSOCIATION,
                    task_reference_constants.MANIFEST_SECTION_NAMES: dict(),
                    task_reference_constants.MANIFEST_ITEM_NAMES: dict(),
                    task_reference_constants.MANIFEST_ACCOUNT_IDS: dict(),
                }
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_puppet_association_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # GET THE SPOKE PRODUCTS AND VERSIONS SO WE KNOW WHAT IS MISSING OR NEEDS UPDATING
            spoke_portfolio_all_products_and_versions_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_all_products_and_versions_ref):
                all_tasks[spoke_portfolio_all_products_and_versions_ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        [spoke_portfolio_puppet_association_ref],
                    ),
                    task_reference=spoke_portfolio_all_products_and_versions_ref,
                    section_name=constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                )
                all_tasks[spoke_portfolio_all_products_and_versions_ref][
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_ref][
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_ref][
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ] = dict()
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))

            # COPY OR IMPORT THE CHANGES BETWEEN THE TWO PORTFOLIOS
            product_generation_method = task_to_add.get("product_generation_method")
            portfolio_import_or_copy_ref = f"portfolio_{product_generation_method}-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"

            portfolio_import_or_copy_task = dict(
                **get_spoke_local_portfolio_common_args(
                    task_to_add,
                    all_tasks_task_reference,
                    [
                        spoke_portfolio_all_products_and_versions_ref,
                        hub_portfolio_all_products_and_versions_before_ref,
                    ],
                ),
                task_reference=portfolio_import_or_copy_ref,
                product_generation_method=product_generation_method,
                section_name=f"portfolio-{product_generation_method}",
                portfolio_get_all_products_and_their_versions_ref=spoke_portfolio_all_products_and_versions_ref,
                portfolio_get_all_products_and_their_versions_for_hub_ref=hub_portfolio_all_products_and_versions_before_ref,
            )
            portfolio_import_or_copy_task[
                task_reference_constants.MANIFEST_SECTION_NAMES
            ] = dict(**task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            portfolio_import_or_copy_task[
                task_reference_constants.MANIFEST_ITEM_NAMES
            ] = dict(**task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            portfolio_import_or_copy_task[
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ] = dict(**task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
            if product_generation_method == constants.PRODUCT_GENERATION_METHOD_IMPORT:
                portfolio_import_or_copy_task["hub_portfolio_task_reference"] = (
                    hub_portfolio_ref
                )
                portfolio_import_or_copy_task["dependencies_by_reference"].append(
                    hub_portfolio_ref
                )
            all_tasks[portfolio_import_or_copy_ref] = portfolio_import_or_copy_task
            dependencies_for_constraints.append(portfolio_import_or_copy_ref)

        if task_to_add.get("associations"):
            v1_stack_name = f"associations-for-{utils.slugify_for_cloudformation_stack_name(item_name)}"
            task_ref_for_v1_stack = f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-{v1_stack_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            if not all_tasks.get(task_ref_for_v1_stack):
                all_tasks[task_ref_for_v1_stack] = dict(
                    account_id=task_to_add.get("account_id"),
                    region=task_to_add.get("region"),
                    execution=task_to_add.get("execution"),
                    task_reference=task_ref_for_v1_stack,
                    section_name=constants.TERMINATE_CLOUDFORMATION_STACK_TASK,
                    dependencies_by_reference=[],
                    stack_name=v1_stack_name,
                )

            v2_stack_name = (
                association_utils.generate_stack_name_for_associations_by_item_name(
                    item_name
                )
            )
            task_ref_for_v2_stack = f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-{v2_stack_name}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            if not all_tasks.get(task_ref_for_v2_stack):
                all_tasks[task_ref_for_v2_stack] = dict(
                    account_id=task_to_add.get("account_id"),
                    region=task_to_add.get("region"),
                    execution=task_to_add.get("execution"),
                    task_reference=task_ref_for_v2_stack,
                    section_name=constants.TERMINATE_CLOUDFORMATION_STACK_TASK,
                    dependencies_by_reference=[],
                    stack_name=v2_stack_name,
                )

            ref = f"portfolio_associations-{task_to_add.get('portfolio')}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            if all_tasks.get(ref):
                for association in task_to_add.get("associations"):
                    if association not in all_tasks[ref]["associations"]:
                        all_tasks[ref]["associations"].append(association)
                        all_tasks[ref]["dependencies_by_reference"].append(
                            task_ref_for_v1_stack
                        )
                        all_tasks[ref]["dependencies_by_reference"].append(
                            task_ref_for_v2_stack
                        )
            else:
                all_tasks[ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        [
                            constants.CREATE_POLICIES,
                            task_ref_for_v1_stack,
                            task_ref_for_v2_stack,
                        ],
                    ),
                    task_reference=ref,
                    spoke_local_portfolio_name=item_name,
                    section_name=constants.PORTFOLIO_ASSOCIATIONS,
                    associations=task_to_add.get("associations"),
                )
            all_tasks[ref][task_reference_constants.MANIFEST_SECTION_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ITEM_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            all_tasks[ref][task_reference_constants.MANIFEST_ACCOUNT_IDS] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )
        if not is_sharing_with_puppet_account:
            # GET NEW PRODUCT AND VERSIONS FOLLOWING THE IMPORT OR COPY
            spoke_portfolio_all_products_and_versions_after_ref = f"{constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS}-after-{task_to_add.get('account_id')}-{task_to_add.get('region')}-{task_to_add.get('portfolio')}"
            if not all_tasks.get(spoke_portfolio_all_products_and_versions_after_ref):
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        [
                            spoke_portfolio_puppet_association_ref,
                            portfolio_import_or_copy_ref,
                        ],
                    ),
                    task_reference=spoke_portfolio_all_products_and_versions_after_ref,
                    section_name=constants.PORTFOLIO_GET_ALL_PRODUCTS_AND_THEIR_VERSIONS,
                )
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_SECTION_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_ITEM_NAMES
                ] = dict()
                all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                    task_reference_constants.MANIFEST_ACCOUNT_IDS
                ] = dict()
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_SECTION_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_ITEM_NAMES
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES))
            all_tasks[spoke_portfolio_all_products_and_versions_after_ref][
                task_reference_constants.MANIFEST_ACCOUNT_IDS
            ].update(task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS))
            dependencies_for_constraints.append(
                spoke_portfolio_all_products_and_versions_after_ref,
            )

        shared_ref = f"{task_to_add.get('portfolio')}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
        if task_to_add.get("launch_constraints"):
            stack_to_terminate = f"launch-constraints-for-{utils.slugify_for_cloudformation_stack_name(item_name)}"
            task_ref_for_stack_to_terminate = f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-{stack_to_terminate}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            if not all_tasks.get(task_ref_for_stack_to_terminate):
                all_tasks[task_ref_for_stack_to_terminate] = dict(
                    account_id=task_to_add.get("account_id"),
                    region=task_to_add.get("region"),
                    execution=task_to_add.get("execution"),
                    task_reference=task_ref_for_stack_to_terminate,
                    section_name=constants.TERMINATE_CLOUDFORMATION_STACK_TASK,
                    dependencies_by_reference=[],
                    stack_name=stack_to_terminate,
                )

            ref = f"launch_constraints-{shared_ref}"
            if all_tasks.get(ref):
                task = all_tasks[ref]
                if task.get("execution") != task_to_add.get("execution"):
                    raise Exception(
                        f"{item_name} and another is setting launch_constraints for the same portfolio using different execution modes"
                    )
                task["dependencies_by_reference"] = (
                    task["dependencies_by_reference"]
                    + [all_tasks_task_reference]
                    + dependencies_for_constraints
                )

            else:
                task = all_tasks[ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        dependencies_for_constraints,
                    ),
                    task_reference=ref,
                    section_name=constants.PORTFOLIO_CONSTRAINTS_LAUNCH,
                    spoke_local_portfolio_name=item_name,
                    launch_constraints=dict(products=dict()),
                    portfolio_get_all_products_and_their_versions_ref=spoke_portfolio_all_products_and_versions_after_ref,
                )
            all_tasks[ref]["dependencies_by_reference"].append(
                task_ref_for_stack_to_terminate
            )

            for launch_constraint in task_to_add["launch_constraints"]:
                if launch_constraint.get("product"):
                    product = launch_constraint.get("product")
                    if not task["launch_constraints"]["products"].get(product):
                        task["launch_constraints"]["products"][product] = []

                    for role in launch_constraint.get("roles"):
                        if role not in task["launch_constraints"]["products"][product]:
                            task["launch_constraints"]["products"][product].append(role)
                if launch_constraint.get("products"):
                    for product in launch_constraint["products"]:
                        if not task["launch_constraints"]["products"].get(product):
                            task["launch_constraints"]["products"][product] = []
                        for role in launch_constraint.get("roles"):
                            if (
                                role
                                not in task["launch_constraints"]["products"][product]
                            ):
                                task["launch_constraints"]["products"][product].append(
                                    role
                                )

            task[task_reference_constants.MANIFEST_SECTION_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            task[task_reference_constants.MANIFEST_ITEM_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            task[task_reference_constants.MANIFEST_ACCOUNT_IDS] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )

        if task_to_add.get("resource_update_constraints"):
            stack_to_terminate = f"update-resource-constraints-for-{utils.slugify_for_cloudformation_stack_name(item_name)}"
            task_ref_for_stack_to_terminate = f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-{stack_to_terminate}-{task_to_add.get('account_id')}-{task_to_add.get('region')}"
            if not all_tasks.get(task_ref_for_stack_to_terminate):
                all_tasks[task_ref_for_stack_to_terminate] = dict(
                    account_id=task_to_add.get("account_id"),
                    region=task_to_add.get("region"),
                    execution=task_to_add.get("execution"),
                    task_reference=task_ref_for_stack_to_terminate,
                    section_name=constants.TERMINATE_CLOUDFORMATION_STACK_TASK,
                    dependencies_by_reference=[],
                    stack_name=stack_to_terminate,
                )

            ref = f"resource_update_constraints-{shared_ref}"
            if all_tasks.get(ref):
                task = all_tasks[ref]
                if task.get("execution") != task_to_add.get("execution"):
                    raise Exception(
                        f"{item_name} and another is setting resource_update_constraints for the same portfolio using different execution modes"
                    )
                task["dependencies_by_reference"] = (
                    task["dependencies_by_reference"]
                    + [all_tasks_task_reference]
                    + dependencies_for_constraints
                )

            else:
                task = all_tasks[ref] = dict(
                    **get_spoke_local_portfolio_common_args(
                        task_to_add,
                        all_tasks_task_reference,
                        dependencies_for_constraints,
                    ),
                    task_reference=ref,
                    section_name=constants.PORTFOLIO_CONSTRAINTS_RESOURCE_UPDATE,
                    spoke_local_portfolio_name=item_name,
                    resource_update_constraints=dict(products=dict()),
                    portfolio_get_all_products_and_their_versions_ref=spoke_portfolio_all_products_and_versions_after_ref,
                )
            all_tasks[ref]["dependencies_by_reference"].append(
                task_ref_for_stack_to_terminate
            )

            for resource_update_constraint in task_to_add[
                "resource_update_constraints"
            ]:
                if resource_update_constraint.get("product"):
                    product = resource_update_constraint.get("product")
                    if task["resource_update_constraints"]["products"].get(
                        product
                    ) and task["resource_update_constraints"]["products"][
                        product
                    ] != resource_update_constraint.get(
                        "tag_update_on_provisioned_product"
                    ):
                        raise Exception(
                            f"{section_name} {item_name} is trying to override the resource_update_constraints"
                        )
                    task["resource_update_constraints"]["products"][product] = (
                        resource_update_constraint.get(
                            "tag_update_on_provisioned_product"
                        )
                    )
                if resource_update_constraint.get("products"):
                    for product in resource_update_constraint["products"]:
                        if task["resource_update_constraints"]["products"].get(
                            product
                        ) and task["resource_update_constraints"]["products"][
                            product
                        ] != resource_update_constraint.get(
                            "tag_update_on_provisioned_product"
                        ):
                            raise Exception(
                                f"{section_name} {item_name} is trying to override the resource_update_constraints"
                            )
                        task["resource_update_constraints"]["products"][product] = (
                            resource_update_constraint.get(
                                "tag_update_on_provisioned_product"
                            )
                        )

            task[task_reference_constants.MANIFEST_SECTION_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_SECTION_NAMES)
            )
            task[task_reference_constants.MANIFEST_ITEM_NAMES] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ITEM_NAMES)
            )
            task[task_reference_constants.MANIFEST_ACCOUNT_IDS] = dict(
                **task_to_add.get(task_reference_constants.MANIFEST_ACCOUNT_IDS)
            )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

CONTENT = """<!DOCTYPE HTML>
<html>
<head>
    <title>Timeline</title>

    <style type="text/css">
        body, html {{
            font-family: sans-serif;
        }}
        .success {{
            color: white !important;
        }}
        .failure {{
            color: red !important;
        }}
        .AppSectionTask {{
            background-color: #ffdd00 !important;
        }}
        .AssertionsSectionTask {{
            background-color: #00ffe1 !important;
        }}
        .CodeBuildRunsSectionTask {{
            background-color: rgb(125, 41, 175) !important;
        }}
        .GenericScheduleRunDeployInSpokeTask {{
            background-color: #646164 !important;
        }}
        .GetCloudFormationTemplateFromS3 {{
            background-color: #646164 !important;
        }}
        .GetSSMParamIndividuallyTask {{
            background-color: #9b82fa !important;
        }}
        .GetSSMParamTask {{
            background-color: #9b82fa !important;
        }}
        .LambdaInvocationsSectionTask {{
            background-color: #a6e391 !important;
        }}
        .LaunchSectionTask {{
            background-color: #ff7300 !important;
        }}
        .ProvisionStackTask {{
            background-color: #f397a3 !important;
        }}
        .ServiceControlPoliciesSectionTask {{
            background-color: #2cadf6 !important;
        }}
        .SimulatePolicysSectionTask {{
            background-color: #8c6e00 !important;
        }}
        .SpokeLocalPortfolioSectionTask {{
            background-color: #005586 !important;
        }}
        .StackSectionTask {{
            background-color: #f397a3 !important;
        }}
        .StackTask {{
            background-color: #f397a3 !important;
        }}
        .TagPoliciesSectionTask {{
            background-color: #2cadf6 !important;
        }}
        .WorkspaceSectionTask {{
            background-color: #137206 !important;
        }}

    </style>

    <script src="https://visjs.github.io/vis-timeline/standalone/umd/vis-timeline-graph2d.min.js" rel="script"></script>
    <link href="https://visjs.github.io/vis-timeline/styles/vis-timeline-graph2d.min.css" rel="stylesheet"
          type="text/css"/>
</head>
<body>
<div id="visualization"></div>

<script type="text/javascript">
    // DOM element where the Timeline will be attached
    var container = document.getElementById('visualization');

    // Create a DataSet (allows two way data-binding)
    var items = new vis.DataSet({DATASET});
    
    var groups = new vis.DataSet({GROUPS});

    // Configuration for the Timeline
    var options = {{
    
        start: "{START}",
        end: "{END}",

        height: 800,

        selectable: false,

        showTooltips: true,

        tooltip: {{
            followMouse: true,
        }},

        timeAxis: {{scale: 'second', step: 10}},

        format: {{
            minorLabels: function (date, scale, step) {{
                return date.format("ss");
            }},
            majorLabels: function (date, scale, step) {{
                return date.format("HH:mm:ss");
            }}
        }},

    }};

    // Create a Timeline    
    var timeline = new vis.Timeline({PARAMS});
</script>
</body>
</html>"""



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


sharing_mode_map = dict(
    ACCOUNT="IMPORTED",
    AWS_ORGANIZATIONS="AWS_ORGANIZATIONS",
)


class GetPortfolioLocalTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    status = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def get_portfolio_details(self):
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            paginator = servicecatalog.get_paginator("list_portfolios")
            for page in paginator.paginate():
                for portfolio_details in page.get("PortfolioDetails", []):
                    if portfolio_details.get("DisplayName") == self.portfolio:
                        return portfolio_details
        if self.status == constants.TERMINATED:
            return {}
        else:
            raise Exception(f"Could not find portfolio: {self.portfolio}")

    def run(self):
        self.write_output(self.get_portfolio_details())


class GetPortfolioImportedTask(tasks.TaskWithReference):
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    sharing_mode = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "portfolio": self.portfolio,
            "sharing_mode": self.sharing_mode,
            "region": self.region,
            "account_id": self.account_id,
        }

    def get_portfolio_details(self):
        mode = sharing_mode_map.get(self.sharing_mode)
        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            paginator = servicecatalog.get_paginator("list_accepted_portfolio_shares")
            for page in paginator.paginate(
                PortfolioShareType=mode,
            ):
                for portfolio_details in page.get("PortfolioDetails", []):
                    if portfolio_details.get("DisplayName") == self.portfolio:
                        return portfolio_details

        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            paginator = servicecatalog.get_paginator("list_accepted_portfolio_shares")
            for page in paginator.paginate(
                PortfolioShareType=(
                    "IMPORTED" if mode == "AWS_ORGANIZATIONS" else "AWS_ORGANIZATIONS"
                ),
            ):
                for portfolio_details in page.get("PortfolioDetails", []):
                    if portfolio_details.get("DisplayName") == self.portfolio:
                        return portfolio_details

        raise Exception(f"Could not find portfolio: {self.portfolio}")

    def run(self):
        self.write_output(self.get_portfolio_details())



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class RunDeployInSpokeTaskTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    task_reference = "task_reference"
    dependencies_by_reference = []
    account_id = "account_id"
    generate_manifest_ref = "generate_manifest_ref"

    def setUp(self) -> None:
        import os
        from servicecatalog_puppet import environmental_variables

        os.environ[environmental_variables.REGIONS] = "[]"
        from servicecatalog_puppet.workflow.launch import run_deploy_in_spoke_task

        self.module = run_deploy_in_spoke_task

        self.sut = self.module.RunDeployInSpokeTask(
            **self.get_common_args(),
            account_id=self.account_id,
            generate_manifest_ref=self.generate_manifest_ref,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "puppet_account_id": self.puppet_account_id,
            "account_id": self.account_id,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import unittest

from servicecatalog_puppet import constants, task_reference_constants


class ImportedPortfoliosTest(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        from servicecatalog_puppet.commands.task_reference_helpers.generators import (
            spoke_local_portfolios,
        )

        self.sut = spoke_local_portfolios

    def test_for_accounts(self):
        # setup
        puppet_account_id = "hub_account_id"
        account_id = "spoke_account_id"
        region = "eu-west-0"
        ou_name = "ou-do8d-me7f39on"
        item_name = "depsrefactor"
        portfolio = "DepsRefactor"

        section_name = constants.IMPORTED_PORTFOLIOS
        task_to_add = {
            "spoke_local_portfolio_name": item_name,
            "execution": "hub",
            "sharing_mode": "AWS_ORGANIZATIONS",
            "share_tag_options": "True",
            "share_principals": "True",
            "associations": ["arn:aws:iam::${AWS::AccountId}:role/Admin"],
            "portfolio": portfolio,
            "puppet_account_id": puppet_account_id,
            "status": None,
            "requested_priority": 0,
            "dependencies": [],
            "account_id": account_id,
            "organization": "o-sw3edla4pd",
            "ou": ou_name,
            "region": region,
            task_reference_constants.MANIFEST_SECTION_NAMES: {
                "imported-portfolios": True
            },
            task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
            task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
            "section_name": section_name,
            "item_name": item_name,
            "dependencies_by_reference": ["create-policies"],
            "task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
            "resources_required": [
                f"SERVICE_CATALOG_LIST_PORTFOLIOS_{region}_OF_{account_id}",
                f"SERVICE_CATALOG_CREATE_PORTFOLIOS_{region}_OF_{account_id}",
            ],
            "product_generation_method": "import",
        }
        all_tasks_task_reference = task_to_add["task_reference"]
        all_tasks = {all_tasks_task_reference: task_to_add}
        task_reference = f"{account_id}-{region}"

        # exercise
        self.sut.handle_spoke_local_portfolios(
            all_tasks,
            all_tasks_task_reference,
            item_name,
            puppet_account_id,
            section_name,
            task_reference,
            task_to_add,
        )

        # verify
        n_all_tasks = len(all_tasks.keys())
        self.assertEqual(task_to_add, all_tasks[all_tasks_task_reference])
        self.assertEqual(
            {
                "account_id": puppet_account_id,
                "dependencies_by_reference": [],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "puppet_account_id": puppet_account_id,
                "region": region,
                "section_name": constants.PORTFOLIO_LOCAL,
                "status": None,
                "task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
            },
            all_tasks[f"portfolio-local-{puppet_account_id}-{region}-{portfolio}"],
        )

        self.assertEqual(
            {
                "account_id": puppet_account_id,
                "dependencies_by_reference": [
                    f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                    "create-policies",
                ],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "section_name": "portfolio-puppet-role-association",
                "task_reference": f"portfolio-puppet-role-association-{puppet_account_id}-{region}-{portfolio}",
            },
            all_tasks[
                f"portfolio-puppet-role-association-{puppet_account_id}-{region}-{portfolio}"
            ],
        )

        self.assertEqual(
            {
                "account_id": puppet_account_id,
                "dependencies_by_reference": [
                    f"portfolio-local-{puppet_account_id}-{region}-{portfolio}"
                ],
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio_task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "region": region,
                "section_name": "describe-portfolio-shares",
                "task_reference": f"describe-portfolio-shares-ORGANIZATIONAL_UNIT-portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "type": "ORGANIZATIONAL_UNIT",
            },
            all_tasks[
                f"describe-portfolio-shares-ORGANIZATIONAL_UNIT-portfolio-local-{puppet_account_id}-{region}-{portfolio}"
            ],
        )

        self.assertEqual(
            {
                "account_id": account_id,
                "dependencies_by_reference": [
                    f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                    "create-policies",
                    f"describe-portfolio-shares-ORGANIZATIONAL_UNIT-portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                ],
                "describe_portfolio_shares_task_ref": f"describe-portfolio-shares-ORGANIZATIONAL_UNIT-portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "ou_to_share_with": ou_name,
                "portfolio": portfolio,
                "portfolio_task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "section_name": "portfolio-share-and-accept-aws_organizations",
                "share_principals": "True",
                "share_tag_options": "True",
                "task_reference": f"portfolio_share_and_accept-{ou_name}-{region}-{portfolio}",
            },
            all_tasks[f"portfolio_share_and_accept-{ou_name}-{region}-{portfolio}"],
        )

        self.assertEqual(
            {
                "account_id": puppet_account_id,
                "dependencies_by_reference": [
                    f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                    f"portfolio-puppet-role-association-{puppet_account_id}-{region}-{portfolio}",
                ],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio_task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "section_name": "portfolio-get-all-products-and-their-versions",
                "task_reference": f"portfolio-get-all-products-and-their-versions-before-{puppet_account_id}-{region}-{portfolio}",
            },
            all_tasks[
                f"portfolio-get-all-products-and-their-versions-before-{puppet_account_id}-{region}-{portfolio}"
            ],
        )

        self.assertEqual(
            {
                "account_id": account_id,
                "dependencies_by_reference": [
                    f"imported-portfolios_{item_name}_{account_id}_{region}",
                    "create-policies",
                    f"portfolio_share_and_accept-{ou_name}-{region}-{portfolio}",
                ],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
                "puppet_account_id": puppet_account_id,
                "region": region,
                "section_name": "portfolio-puppet-role-association",
                "task_reference": f"portfolio-puppet-role-association-{account_id}-{region}-{portfolio}",
            },
            all_tasks[
                f"portfolio-puppet-role-association-{account_id}-{region}-{portfolio}"
            ],
        )

        self.assertEqual(
            {
                "account_id": account_id,
                "dependencies_by_reference": [
                    f"imported-portfolios_{item_name}_{account_id}_{region}",
                    f"portfolio-puppet-role-association-{account_id}-{region}-{portfolio}",
                ],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
                "region": region,
                "section_name": "portfolio-get-all-products-and-their-versions",
                "status": None,
                "task_reference": f"portfolio-get-all-products-and-their-versions-{account_id}-{region}-{portfolio}",
            },
            all_tasks[
                f"portfolio-get-all-products-and-their-versions-{account_id}-{region}-{portfolio}"
            ],
        )

        t = all_tasks[f"portfolio_import-{account_id}-{region}-{portfolio}"]
        self.assertEqual(
            {
                "account_id": account_id,
                "dependencies_by_reference": [
                    f"imported-portfolios_{item_name}_{account_id}_{region}",
                    f"portfolio-get-all-products-and-their-versions-{account_id}-{region}-{portfolio}",
                    f"portfolio-get-all-products-and-their-versions-before-{puppet_account_id}-{region}-{portfolio}",
                    f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                ],
                "execution": "hub",
                "hub_portfolio_task_reference": f"portfolio-local-{puppet_account_id}-{region}-{portfolio}",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_get_all_products_and_their_versions_for_hub_ref": f"portfolio-get-all-products-and-their-versions-before-{puppet_account_id}-{region}-{portfolio}",
                "portfolio_get_all_products_and_their_versions_ref": f"portfolio-get-all-products-and-their-versions-{account_id}-{region}-{portfolio}",
                "portfolio_task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
                "product_generation_method": "import",
                "region": region,
                "section_name": "portfolio-import",
                "status": None,
                "task_reference": f"portfolio_import-{account_id}-{region}-{portfolio}",
            },
            t,
        )

        self.assertEqual(
            {
                "account_id": account_id,
                "associations": ["arn:aws:iam::${AWS::AccountId}:role/Admin"],
                "dependencies_by_reference": [
                    f"imported-portfolios_{item_name}_{account_id}_{region}",
                    "create-policies",
                    f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-associations-for-{item_name}-{account_id}-{region}",
                    f"{constants.TERMINATE_CLOUDFORMATION_STACK_TASK}-associations-v2-for-{item_name}-{account_id}-{region}",
                ],
                "execution": "hub",
                "spoke_local_portfolio_name": item_name,
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
                "region": region,
                "section_name": "portfolio-associations",
                "status": None,
                "task_reference": f"portfolio_associations-{portfolio}-{account_id}-{region}",
            },
            all_tasks[f"portfolio_associations-{portfolio}-{account_id}-{region}"],
        )

        self.assertEqual(
            {
                "account_id": account_id,
                "dependencies_by_reference": [
                    f"imported-portfolios_{item_name}_{account_id}_{region}",
                    f"portfolio-puppet-role-association-{account_id}-{region}-{portfolio}",
                    f"portfolio_import-{account_id}-{region}-{portfolio}",
                ],
                "execution": "hub",
                task_reference_constants.MANIFEST_ACCOUNT_IDS: {account_id: True},
                task_reference_constants.MANIFEST_ITEM_NAMES: {item_name: True},
                task_reference_constants.MANIFEST_SECTION_NAMES: {
                    "imported-portfolios": True
                },
                "portfolio": portfolio,
                "portfolio_task_reference": f"imported-portfolios_{item_name}_{account_id}_{region}",
                "region": region,
                "section_name": f"portfolio-get-all-products-and-their-versions",
                "status": None,
                "task_reference": f"portfolio-get-all-products-and-their-versions-after-{account_id}-{region}-{portfolio}",
            },
            all_tasks[
                f"portfolio-get-all-products-and-their-versions-after-{account_id}-{region}-{portfolio}"
            ],
        )
        self.assertEqual(13, n_all_tasks)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import functools

from betterboto import client as betterboto_client

from servicecatalog_puppet import config


@functools.lru_cache(maxsize=10)
def get_cache_download_client():
    return betterboto_client.CrossAccountClientContextManager(
        "s3",
        config.get_cache_download_role_arn(config.get_executor_account_id()),
        "s3-client",
    )



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

from unittest import skip

from servicecatalog_puppet.workflow import tasks_unit_tests_helper


class SSMOutputsTasksTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    account_id = "account_id"
    region = "region"
    param_name = "param_name"
    stack_output = "stack_output"
    task_generating_output = "task_generating_output"
    task_generating_output_account_id = "task_generating_output_account_id"
    task_generating_output_region = "task_generating_output_region"
    task_generating_output_section_name = "task_generating_output_section_name"
    task_generating_output_entity_name = "task_generating_output_entity_name"
    task_generating_output_stack_set_name = "task_generating_output_stack_set_name"
    task_generating_output_launch_name = "task_generating_output_launch_name"
    force_operation = False

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.ssm import ssm_outputs_task

        self.module = ssm_outputs_task

        self.sut = self.module.SSMOutputsTasks(
            **self.get_common_args(),
            account_id=self.account_id,
            region=self.region,
            param_name=self.param_name,
            stack_output=self.stack_output,
            task_generating_output=self.task_generating_output,
            task_generating_output_account_id=self.task_generating_output_account_id,
            task_generating_output_region=self.task_generating_output_region,
            task_generating_output_section_name=self.task_generating_output_section_name,
            task_generating_output_entity_name=self.task_generating_output_entity_name,
            task_generating_output_stack_set_name=self.task_generating_output_stack_set_name,
            task_generating_output_launch_name=self.task_generating_output_launch_name,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "param_name": self.param_name,
            "stack_output": self.stack_output,
            "force_operation": self.force_operation,
            "updated": False,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()


class TerminateSSMOutputsTasksTest(tasks_unit_tests_helper.PuppetTaskUnitTest):
    puppet_account_id = "puppet_account_id"
    account_id = "account_id"
    region = "region"
    param_name = "param_name"

    def setUp(self) -> None:
        from servicecatalog_puppet.workflow.ssm import ssm_outputs_task

        self.module = ssm_outputs_task

        self.sut = self.module.TerminateSSMOutputsTasks(
            **self.get_common_args(),
            account_id=self.account_id,
            region=self.region,
            param_name=self.param_name,
        )

        self.wire_up_mocks()

    def test_params_for_results_display(self):
        # setup
        expected_result = {
            "task_reference": self.task_reference,
            "account_id": self.account_id,
            "region": self.region,
            "param_name": self.param_name,
        }

        # exercise
        actual_result = self.sut.params_for_results_display()

        # verify
        self.assertEqual(expected_result, actual_result)

    @skip
    def test_run(self):
        # setup
        # exercise
        actual_result = self.sut.run()

        # verify
        raise NotImplementedError()



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import luigi

from servicecatalog_puppet import constants
from servicecatalog_puppet.workflow.dependencies import tasks


class DescribePortfolioSharesTask(tasks.TaskWithReferenceAndCommonParameters):
    portfolio_task_reference = luigi.Parameter()
    type = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "task_reference": self.task_reference,
            "region": self.region,
            "account_id": self.account_id,
            "portfolio_task_reference": self.portfolio_task_reference,
            "type": self.type,
        }

    def run(self):
        portfolio_id = self.get_attribute_from_output_from_reference_dependency(
            "Id", self.portfolio_task_reference
        )

        with self.spoke_regional_client("servicecatalog") as servicecatalog:
            has_more = True
            args = dict(
                PortfolioId=portfolio_id,
                Type=self.type,
            )
            shares = dict()
            while has_more:
                result = servicecatalog.describe_portfolio_shares(**args)
                for portfolio_share_detail in result.get("PortfolioShareDetails", []):
                    shares[portfolio_share_detail.get("PrincipalId")] = (
                        portfolio_share_detail
                    )
                if result.get("NextPageToken"):
                    has_more = True
                    args["PageToken"] = result.get("NextPageToken")
                else:
                    has_more = False
            self.write_output(shares)



#  Copyright 2023 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#  SPDX-License-Identifier: Apache-2.0

import re

import luigi
import troposphere as t
from troposphere import servicecatalog

from servicecatalog_puppet import config, constants, utils
from servicecatalog_puppet.workflow.dependencies import tasks


class CreateUpdateResourceConstraintsForSpokeLocalPortfolioTask(
    tasks.TaskWithReference
):
    portfolio_task_reference = luigi.Parameter()
    spoke_local_portfolio_name = luigi.Parameter()
    account_id = luigi.Parameter()
    region = luigi.Parameter()
    portfolio = luigi.Parameter()
    resource_update_constraints = luigi.ListParameter()

    portfolio_get_all_products_and_their_versions_ref = luigi.Parameter()
    cachable_level = constants.CACHE_LEVEL_RUN

    def params_for_results_display(self):
        return {
            "puppet_account_id": self.puppet_account_id,
            "spoke_local_portfolio_name": self.spoke_local_portfolio_name,
            "portfolio": self.portfolio,
            "region": self.region,
            "account_id": self.account_id,
        }

    def run(self):
        portfolio_details = self.get_output_from_reference_dependency(
            self.portfolio_task_reference
        )
        portfolio_id = portfolio_details.get("Id")
        products_and_their_versions = self.get_output_from_reference_dependency(
            self.portfolio_get_all_products_and_their_versions_ref
        )

        tpl = t.Template()
        tpl.description = f"update resource constraints for {self.portfolio}"

        for (
            product_to_constrain,
            validate_tag_update,
        ) in self.resource_update_constraints.get("products", {}).items():
            found = False
            for (
                product_name,
                product_details,
            ) in products_and_their_versions.items():
                if re.match(
                    product_to_constrain,
                    product_name,
                ):
                    found = True
                    product_id = product_details.get("ProductId")

                    tpl.add_resource(
                        servicecatalog.ResourceUpdateConstraint(
                            f"L{ portfolio_id}B{product_id}C".replace("-", ""),
                            PortfolioId=portfolio_id,
                            Description=f"TagUpdate = {validate_tag_update}",
                            ProductId=product_id,
                            TagUpdateOnProvisionedProduct=validate_tag_update,
                        )
                    )
            if not found:
                raise Exception(
                    f"Did not find a matching product for: {product_to_constrain}"
                )

        template_to_use = tpl.to_yaml()

        with self.spoke_regional_client("cloudformation") as cloudformation:
            stack_name = f"update-resource-constraints-for-{utils.slugify_for_cloudformation_stack_name(self.portfolio)}"
            cloudformation.create_or_update(
                StackName=stack_name,
                TemplateBody=template_to_use,
                NotificationARNs=(
                    [
                        f"arn:{config.get_partition()}:sns:{self.region}:{self.puppet_account_id}:servicecatalog-puppet-cloudformation-regional-events"
                    ]
                    if self.should_use_sns
                    else []
                ),
                ShouldDeleteRollbackComplete=self.should_delete_rollback_complete_stacks,
                Tags=self.initialiser_stack_tags,
            )
            result = cloudformation.describe_stacks(
                StackName=stack_name,
            ).get(
                "Stacks"
            )[0]
            self.write_empty_output()
